(ns com.aurellem.gb-driver
  (:import com.aurellem.gb.Gb)
  (:import java.io.File)
  (:import (java.nio IntBuffer ByteOrder)))

(Gb/loadVBA)

(def yellow-rom-image
  (File. "/home/r/proj/pokemon-escape/roms/yellow.gbc"))

(def yellow-save-file
  (File. "/home/r/proj/pokemon-escape/roms/yellow.sav"))

(def current-frame (atom 0))

(defn vba-init []
  (reset! current-frame 0)
  (.delete yellow-save-file)
  (Gb/startEmulator (.getCanonicalPath yellow-rom-image)))

(defn shutdown [] (Gb/shutdown))

(defn reset [] (shutdown) (vba-init))

(defn cpu-data [size arr-fn]
  (let [store (int-array size)]
    (fn [] 
      (arr-fn store)
      store)))

(def ram
  (cpu-data (Gb/getRAMSize) #(Gb/getRAM %)))

(def rom 
  (cpu-data (Gb/getROMSize) #(Gb/getROM %)))

(def working-ram 
  (cpu-data Gb/WRAM_SIZE #(Gb/getWRAM %)))

(def video-ram 
  (cpu-data Gb/VRAM_SIZE #(Gb/getVRAM %)))

(def registers
  (cpu-data Gb/NUM_REGISTERS #(Gb/getRegisters %)))

(def button-code
  {;; main buttons
   :a         0x0001
   :b         0x0002

   ;; directional pad
   :r         0x0010
   :l         0x0020
   :u         0x0040
   :d         0x0080

   ;; meta buttons
   :select    0x0004
   :start     0x0008

   ;; hard reset -- not really a button
   :reset   0x0800})

(defn button-mask [buttons]
  (reduce bit-or 0x0000 (map button-code buttons)))

(defn buttons [mask]
  (loop [buttons []
         masks (seq button-code)]
    (if (empty? masks) buttons
        (let [[button value] (first masks)]
          (if (not= 0x0000 (bit-and value mask))
            (recur (conj buttons button) (rest masks))
            (recur buttons (rest masks)))))))


(defn save-state [] (Gb/saveState))

(def history (atom {}))

(defn goto [frame]
  (let [save (@history frame)]
    (if (not (nil? save))
      (do
        (reset! current-frame frame)
        (Gb/loadState save))
      (println "no backup state"))))

(defn clear-history [] (reset! history {}))

(defn rewind
  ([n] (goto (- @current-frame n)))
  ([] (rewind 1)))
  
(defn backup-state [frame]
  (swap! history #(assoc % frame (save-state))))

(def ^:dynamic *save-history* true)

(defn advance []
  (swap! current-frame inc)
  (if *save-history*
    (let [save (save-state)]
      (backup-state @current-frame))))

(defn step
  ([] (advance) (Gb/step))
  ([mask-or-buttons]
     (advance)
     (if (number? mask-or-buttons)
       (Gb/step mask-or-buttons)
       (Gb/step (button-mask mask-or-buttons)))))

(defn step! [& args]
  (binding [*save-history* false]
    (apply step args)))

(defn frame [] @current-frame)
  

